#include <failsystem.h>
/*
static int fd;
struct sblock sb;
extern pthread_mutex_t fdmux,fatmux;
*/

int fsys_fuse_mknod(const char *path,mode_t modt,dev_t devt);
int fsys_fuse_getdir(const char *dirname, fuse_dirh_t h, fuse_dirfil_t filler);
int fsys_fuse_getattr(const char *path, struct stat *stat);
int fsys_fuse_mkdir(const char *path, mode_t modt);
int fsys_fuse_read(const char *path, char *buf, size_t size, off_t offset, struct fuse_file_info *finfo);
int fsys_fuse_write(const char *path, const char *buf, size_t size, off_t offset,struct fuse_file_info 		*finfo);
int fsys_fuse_open(const char *path, struct fuse_file_info *finfo);
int fsys_fuse_utime (const char *path, struct utimbuf *uti);
int fsys_fuse_flush(const char *path, struct fuse_file_info *finfo);
int fsys_fuse_release(const char *path, struct fuse_file_info *finfo);
int fsys_fuse_unlink(const char *path);
int fsys_fuse_truncate(const char *path, off_t newsize);
int fsys_fuse_rename(const char *path, const char *newpath);
int fsys_fuse_stafs(const char *string, struct statfs *statfs);
int fsys_fuse_chmod(const char *path, mode_t modt);
int fsys_fuse_chown(const char *path, uid_t uid, gid_t gid);
int fsys_mkdentry(const char *path,mode_t modt,dev_t devt,u_int type);
int fsys_fuse_readlink(const char *linkname, char *target, size_t size);
int fsys_fuse_symlink(const char *target, const char *new);


int main (int argc,char **argv){
struct fuse_operations fail_ops;
char **myargv;
int i;
	/*now we force single thread operation */
	
	if (argc < 3){fsys_usage(); exit(1);}
	myargv=(char**)malloc(sizeof(char*) * (argc-1));
	myargv[0]=strdup(argv[0]);
	for (i=2;i<argc;i++){
		myargv[i-1]=strdup(argv[i]);
	}
	myargv[i-1]=strdup("-s");
	
	/*init start*/
	/*fuse-open function gives us a file info struct containing an unsigned long field,
	 *we use it to store a pointer to a local structure, so their size must be equal. */
	if (sizeof(unsigned long)!=sizeof(void*)){fprintf(stderr,"Incompatible ulong and void* types");exit (1);}
	fd=fsys_open(argv[1]);
	if (fd < 0){perror("failsystem.c");fsys_usage();exit(1);}
	fsys_logsopen();
	fsys_sblockget(fd,&sb);

#ifdef FATMAPPED
	fsys_fatload(fd,&sb);
#endif	
	/*initialize function pointer struct*/
	memset((void*)&fail_ops,0,sizeof(struct fuse_operations));
	fail_ops.getattr=fsys_fuse_getattr;
	fail_ops.getdir=fsys_fuse_getdir;
	fail_ops.mknod=fsys_fuse_mknod;
	fail_ops.mkdir=fsys_fuse_mkdir;
	fail_ops.read=fsys_fuse_read;
	fail_ops.write=fsys_fuse_write;
	fail_ops.open=fsys_fuse_open;
	fail_ops.utime=fsys_fuse_utime;	/*SETATTR*/
	fail_ops.flush=fsys_fuse_flush;
	fail_ops.release=fsys_fuse_release;
	fail_ops.unlink=fsys_fuse_unlink;
	fail_ops.rmdir=fsys_fuse_unlink;
	fail_ops.truncate=fsys_fuse_truncate;
	fail_ops.rename=fsys_fuse_rename;
	fail_ops.statfs=fsys_fuse_stafs;
	fail_ops.chmod=fsys_fuse_chmod;
	fail_ops.chown=fsys_fuse_chown;
	fail_ops.readlink=fsys_fuse_readlink;
	fail_ops.symlink=fsys_fuse_symlink;
	fuse_main(argc,myargv,&fail_ops);
	
	fsys_sblocksync(fd,&sb);
	fsys_printlog(FSYS_LOG_SYSLOG,"Detached thread\n");
	fsys_close(fd);
	fsys_logsclose();
	return 0;
}

/*LOOKUP reads a directory content, fill names of directory entries of given father (dirname)*/
int fsys_fuse_getdir(const char *dirname, fuse_dirh_t h, fuse_dirfil_t filler){
struct dentry *father,tempdir,sondir;
u_int brotherscount,i,offset;
	fsys_printlog(FSYS_LOG_SYSLOG,"fsys::getdir\n\n");
	fsys_lookupfile(fd,&sb,(char*)dirname,NULL,NULL,&father,&offset);
	fsys_dentryread(fd,&sb,father,(void*)&tempdir,offset,sizeof(struct dentry));
	
	filler(h,".",0,0);
	filler(h,"..",0,0);
	
	brotherscount=tempdir.stat.st_size / (sizeof(struct dentry));
	for (i=0;i<brotherscount;i++){
		fsys_dentryread(fd,&sb,&tempdir,(void*)&sondir,i*sizeof(struct dentry),sizeof(struct dentry));
		filler(h,sondir.name,0,0);
	}
	free(father);
	return 0;
}

/*creates a node in the filesystem ( regular file)*/
int fsys_fuse_mknod(const char *path,mode_t modt,dev_t devt){
	fsys_printlog(FSYS_LOG_SYSLOG,"fsys::mknod\n\n");
	return fsys_mkdentry(path,modt,devt,FSYS_TYPE_REG);
}

/*creates a directory node in the filesystem (directory file)*/
int fsys_fuse_mkdir(const char *path, mode_t modt){
	fsys_printlog(FSYS_LOG_SYSLOG,"fsys::mkdir\n\n");
	return fsys_mkdentry(path,modt,0,FSYS_TYPE_DIR);
}

/*interface to fuse witch provides mkdir and mknod*/
int fsys_mkdentry(const char *path,mode_t modt,dev_t devt,u_int type){
char *strtemp,*filename;
struct dentry father,*gfather,*tmpdir;
u_int goffset;
int ret;

	fsys_stringsplit(path,&strtemp,&filename);
	fsys_lookupfile(fd,&sb,strtemp,NULL,NULL,&gfather,&goffset);
	free(strtemp);
	/*get father*/
	fsys_dentryread(fd,&sb,gfather,(void*)&father,goffset,sizeof(struct dentry));
	/*create a direntry for a file*/
	
	if (strlen(filename) > FSYS_MAXNAMELENGHT){
	/**/
		free(gfather);
		return (-ENAMETOOLONG);
	}

	tmpdir=fsys_init_dentry(modt,0,filename,type);
	free(filename);
	/*add direntry to father*/
	ret=fsys_dentryadd(fd,&sb,&father,tmpdir);
	if (ret==FSYS_DISKFULL){
		free(gfather);
		free(tmpdir);
		return (-ENOSPC);
	}
	/*modify size of father. gfather and goffset identifies father*/
	fsys_dentrywrite(fd,&sb,gfather,(void*)&father,goffset,sizeof(struct dentry));
	
	free(gfather);
	free(tmpdir);

	fsys_sblocksync(fd,&sb);
	return 0;
}

/*get the attributes of a directory entry*/
int fsys_fuse_getattr(const char *path, struct stat *stat){
int ret;
u_int offset;
struct dentry *father,dir;
	fsys_printlog(FSYS_LOG_SYSLOG,"fsys::gettattr\n\n");

	ret=fsys_lookupfile(fd,&sb,(char*)path,NULL,NULL,&father,&offset);
	if (ret==FSYS_FILEMATCH){
		fsys_dentryread(fd,&sb,father,(void*)&dir,offset,sizeof(struct dentry));
		*stat=dir.stat;
		free(father);
		return 0;
	}
	else{
		return (-ENOENT);
	}
}

/* read regular file retrieved trought a data structure pointed by finfo->fh*/
int fsys_fuse_read(const char *path, char *buf, size_t size, off_t offset, struct fuse_file_info *finfo){
int ret;
struct dentry *father,dir;
u_int foffset;
	fsys_printlog(FSYS_LOG_SYSLOG,"fsys::read\n\n");
	
	father=((struct fhandle*)(finfo->fh))->father;
	foffset=((struct fhandle*)(finfo->fh))->offset;
	ret=FSYS_FILEMATCH;
	if (ret==FSYS_FILEMATCH){
		fsys_dentryread(fd,&sb,father,&dir,foffset,sizeof(struct dentry));/*get dir*/
		return fsys_dentryread(fd,&sb,&dir,(void*)buf,offset,size);
	}
	else{
		return (-ENOENT);
	}
}

/* writes data to a regular file*/
int fsys_fuse_write(const char *path, const char *buf, size_t size, off_t offset,struct fuse_file_info *finfo){
struct dentry *father,dir;
u_int bwritten,ret,foffset;

	father=((struct fhandle*)(finfo->fh))->father;
	foffset=((struct fhandle*)(finfo->fh))->offset;
	fsys_printlog(FSYS_LOG_SYSLOG,"fsys::write handle %lu name %s\n",finfo->fh,father->name);
	fsys_dentryread(fd,&sb,father,&dir,foffset,sizeof(struct dentry));/*get dir*/
	bwritten=fsys_dentrywrite(fd,&sb,&dir,(void*)buf,offset,size);
	if (bwritten==FSYS_DISKFULL) {
		return (-ENOSPC);
	}
	
	if ((offset + size) > dir.stat.st_size){
		dir.stat.st_size=offset + bwritten;
		dir.stat.st_blocks=fsys_divideplus(dir.stat.st_size,sb.blocksize);
	}
	dir.stat.st_atime=dir.stat.st_ctime=dir.stat.st_mtime=time(NULL);
	
	ret=fsys_dentrywrite(fd,&sb,(void*)father,&dir,foffset,sizeof(struct dentry));
	if (ret==FSYS_DISKFULL) {
		return (-ENOSPC);
	}
	
	fsys_printlog(FSYS_LOG_SYSLOG,"fsys::write bytes written %u\n\n",bwritten);
	return bwritten;
}

/* open file
 * note: struct fuse_file_info is a area data allocated by fuselib where fh(file handle) is writable,
 * therefore we use fh to store a pointer to a local file relative structure
 */
int fsys_fuse_open(const char *path, struct fuse_file_info *finfo){
struct dentry *father,dir;
struct fhandle *pfh;
u_int offset;
int aclcheck=0;
	fsys_printlog(FSYS_LOG_SYSLOG,"fsys::open\n\n");
	fsys_lookupfile(fd,&sb,(char*)path,NULL,NULL,&father,&offset);
	fsys_dentryread(fd,&sb,father,(void*)&dir,offset,sizeof(struct dentry));
	aclcheck=fsys_aclcheck(&dir,finfo->flags);
	if(!aclcheck){
		free (father);
		return (-EPERM);
	}
	fsys_printlog(FSYS_LOG_STDERR,"fsys::open permission granted\n");
	pfh=(struct fhandle*)malloc(sizeof(struct fhandle));
	memset(pfh,0,sizeof(struct fhandle));
	pfh->father=father;
	pfh->offset=offset;

	finfo->fh=(unsigned long)pfh;
	return 0;
}

/* change atime/mtime/ctime fields in a directory entry */
int fsys_fuse_utime (const char *path, struct utimbuf *uti){
struct dentry *father,dir;
u_int offset;
	fsys_printlog(FSYS_LOG_SYSLOG,"fsys::utime %s\n\n",path);
	fsys_lookupfile(fd,&sb,(char*)path,NULL,NULL,&father,&offset);
	fsys_dentryread(fd,&sb,father,(void*)&dir,offset,sizeof(struct dentry));
	dir.stat.st_atime=uti->actime;
	dir.stat.st_mtime=uti->modtime;
	dir.stat.st_ctime=uti->modtime;
	fsys_dentrywrite(fd,&sb,father,(void*)&dir,offset,sizeof(struct dentry));
	free(father);
	return 0;
}

/* sync data */
int fsys_fuse_flush(const char *path, struct fuse_file_info *finfo){
	fsys_sblocksync(fd,&sb);
	return 0;
}

/* close a file, here we free the structure initialized in fsys_fuse_open */
int fsys_fuse_release(const char *path, struct fuse_file_info *finfo){
	free (((struct fhandle*)(finfo->fh))->father);
	free ((struct fhandle*)(finfo->fh));
	return 0;
}

/*unlink and rmdir*/
int fsys_fuse_unlink(const char *path){
char *fatherpath,*dirname;
struct dentry  *father,*gfather;
u_int offset,goffset;

	fsys_printlog(FSYS_LOG_SYSLOG,"fsys::unlink %s\n\n",path);
	fsys_stringsplit(path,&fatherpath,&dirname);	
	fsys_lookupfile(fd,&sb,(char*)path,&gfather,&goffset,&father,&offset);
	fsys_dentrydel(fd,&sb,father,dirname,FSYS_TRALL);
	father->stat.st_atime=father->stat.st_mtime=father->stat.st_ctime=time(NULL);
	fsys_dentrywrite(fd,&sb,gfather,(void*)father,goffset,sizeof(struct dentry));
	free(father);
	free(gfather);
	free(fatherpath);
	free(dirname);
	fsys_sblocksync(fd,&sb);
	return 0;
}

/* cut the size of a file */
int fsys_fuse_truncate(const char *path, off_t newsize){
struct dentry *father,tmpdir;
u_int offset;
	fsys_printlog(FSYS_LOG_SYSLOG,"fsys::truncate::size %u",(u_int)newsize);
	
	fsys_lookupfile(fd,&sb,(char*)path,NULL,NULL,&father,&offset);
	fsys_dentryread(fd,&sb,father,(void*)&tmpdir,offset,sizeof(struct dentry));	/*get dir*/
	fsys_dentrytruncate(fd,&sb,&tmpdir,newsize);
	fsys_dentrywrite(fd,&sb,father,(void*)&tmpdir,offset,sizeof(struct dentry));
	free(father);
	fsys_sblocksync(fd,&sb);
	return 0;
}

/* rename a directory entry */
int fsys_fuse_rename(const char *srcpath,const char *dstpath){
struct dentry *sfather, *sgfather,*dfather,*dgfather,srcdir;
u_int soffset, sgoffset,doffset,dgoffset;
char *dfathername,*dname,*sfathername,*sname;
int ret;

	fsys_printlog(FSYS_LOG_SYSLOG,"fsys::rename::source %s destination %s",srcpath,dstpath);
	fsys_stringsplit(srcpath,&sfathername,&sname);
	fsys_lookupfile(fd,&sb,(char*)srcpath,&sgfather,&sgoffset,&sfather,&soffset);/*search src dir*/
	fsys_dentryread(fd,&sb,sfather,(void*)&srcdir,soffset,sizeof(struct dentry));/*get src dir*/
	
	fsys_stringsplit(dstpath,&dfathername,&dname);
	
	if (strlen(dname) > FSYS_MAXNAMELENGHT){
		
		free(sgfather);
		free(sfather);
		
		free(sfathername);
		free(sname);
		free(dfathername);
		free(dname);
		return (-ENAMETOOLONG);
	}
	
	strcpy(srcdir.name,dname);/*change the name to src dir*/
	
	/*search for destination dir*/
	ret=fsys_lookupfile(fd,&sb,(char*)dstpath,&dgfather,&dgoffset,&dfather,&doffset);
	free(dgfather);
	if(ret==FSYS_FILEMATCH){
		/*father exist*/
		fsys_dentrydel(fd,&sb,dfather,dname,FSYS_TRALL);
	}
	else {
		/*get the father, dgfather and dfather are NULL here*/
		ret=fsys_lookupfile(fd,&sb,(char*)dfathername,NULL,NULL,&dgfather,&dgoffset);
		dfather=(struct dentry*)malloc(sizeof(struct dentry));
		fsys_dentryread(fd,&sb,dgfather,(void*)dfather,dgoffset,sizeof(struct dentry));
	}
	
	fsys_dentryadd(fd,&sb,dfather,&srcdir);
	fsys_dentrywrite(fd,&sb,dgfather,(void*)dfather,dgoffset,sizeof(struct dentry));
	
	if(!strcmp(dfathername,sfathername)){
		*sfather=*dfather;
	}
	
	fsys_dentrydel(fd,&sb,sfather,sname,FSYS_TRDENTRY);
	fsys_dentrywrite(fd,&sb,sgfather,(void*)sfather,sgoffset,sizeof(struct dentry));
	fsys_sblocksync(fd,&sb);
	free(sfathername);
	free(sname);
	free(dfathername);
	free(dname);
	free(sgfather);
	free(sfather);
	free(dfather);
	free(dgfather);
	return 0;	
}

/* return file system infos */
int fsys_fuse_stafs(const char *string, struct statfs *statfs){
u_int blocksfree,currentblock;

	memset(statfs,0,sizeof(struct statfs));
	
	blocksfree=0;
	currentblock=sb.firstfree;
	while(currentblock){
		currentblock=fsys_getnextblock(fd,&sb,currentblock);
		blocksfree++;
	}
	fsys_printlog(FSYS_LOG_SYSLOG,"statfs::string %s\n %u",string,blocksfree);
	
	statfs->f_bsize=sb.blocksize;
	statfs->f_blocks=sb.blockscount;
	statfs->f_bfree=blocksfree;
	statfs->f_bavail=blocksfree;
	statfs->f_namelen=FSYS_MAXNAMELENGHT;
	return 0;
}

/* change permissions of directory entry */
int fsys_fuse_chmod(const char *path, mode_t modt){
struct dentry *father,dir;
u_int offset;
	
	fsys_lookupfile(fd,&sb,(char*)path,NULL,NULL,&father,&offset);
	fsys_dentryread(fd,&sb,father,(void*)&dir,offset,sizeof(struct dentry));
	dir.stat.st_mode&=0xFFFFF000;		/*reset permission*/
	dir.stat.st_mode|=(modt & 0x00000FFF);	/*set new permission*/
	
	dir.stat.st_mtime=dir.stat.st_ctime=time(NULL);
	
	fsys_dentrywrite(fd,&sb,father,(void*)&dir,offset,sizeof(struct dentry));
	free(father);
	return 0;
}

/* change owner of directory entry */
int fsys_fuse_chown(const char *path, uid_t uid, gid_t gid){
struct dentry *father,dir;
u_int offset;
	
	fsys_lookupfile(fd,&sb,(char*)path,NULL,NULL,&father,&offset);
	fsys_dentryread(fd,&sb,father,(void*)&dir,offset,sizeof(struct dentry));
	dir.stat.st_uid=uid;
	dir.stat.st_gid=gid;
	
	dir.stat.st_mtime=dir.stat.st_ctime=time(NULL);
	
	fsys_dentrywrite(fd,&sb,father,(void*)&dir,offset,sizeof(struct dentry));
	free(father);
	return 0;
}

/* create a symbolic link */
int fsys_fuse_symlink(const char *target, const char *new){
struct dentry *father,*gfather,*dirlink;
char *newprefix,*newname;
u_int goffset;
	/*create directory entry for the link*/
	fsys_stringsplit(new,&newprefix,&newname);

	if (strlen(newname) > FSYS_MAXNAMELENGHT){
		free(newprefix);
		free(newname);
		return (-ENAMETOOLONG);
	}
	/*create the content of link*/
	dirlink=fsys_init_dentry(0777,0,newname,FSYS_TYPE_SYMLINK);
	free(newname);
	fsys_dentrywrite(fd,&sb,dirlink,(void*)target,0,strlen(target)+1);
	dirlink->stat.st_size=strlen(target)+1;
	/*add the link direntry to father*/
	fsys_lookupfile(fd,&sb,newprefix,NULL,NULL,&gfather,&goffset);
	free(newprefix);
	father=(struct dentry*)malloc(sizeof(struct dentry));
	fsys_dentryread(fd,&sb,gfather,(void*)father,goffset,sizeof(struct dentry));
	if (fsys_dentryadd(fd,&sb,father,dirlink) == FSYS_DISKFULL){
		free(dirlink);
		free(father);
		free(gfather);
		return (-ENOSPC);
	}
	free(dirlink);
	fsys_dentrywrite(fd,&sb,gfather,(void*)father,goffset,sizeof(struct dentry));
	free(father);
	free(gfather);
	fsys_sblocksync(fd,&sb);
	return 0;
}

/* resolve a symbolic link */
int fsys_fuse_readlink(const char *linkname, char *target, size_t size){
struct dentry *father,linkdir;
u_int offset;
	
	fsys_lookupfile(fd,&sb,(char*)linkname,NULL,NULL,&father,&offset);
	fsys_dentryread(fd,&sb,father,(void*)&linkdir,offset,sizeof(struct dentry));
	fsys_dentryread(fd,&sb,&linkdir,(void*)target,0,linkdir.stat.st_size);
	free(father);
	return 0;
}
