#ifndef __FAILSYSTEM_H__
#define __FAILSYSTEM_H__

#define FATMAPPED

/* uncomment here for pthread support	
 * #define MTHREAD	
*/
#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include <fuse.h>
#include <errno.h>
#include <time.h>
#include <stdarg.h>	/*va_list*/
#include <pthread.h>	/*pthread*/
#include <string.h>	/*memset*/
#include <syslog.h>

/************************************ CONSTANTS DEFINITIONS ***********************************/
#define FSYS_MAXNAMELENGHT 32

/*those defines directory entry type, conform to dirent.h*/
#define	FSYS_TYPE_DIR		4
#define	FSYS_TYPE_REG		8
#define FSYS_TYPE_SYMLINK	10
	
#define FSYS_TRALL	1
#define FSYS_TRDENTRY	0

#define FSYS_RESERVED	0
#define FSYS_DISKFULL	0
#define FSYS_END	0
#define FSYS_BUSY	0
#define FSYS_FILEMATCH 0
#define FSYS_FILENOMATCH (-1)
#define FSYS_NAME_OVERFLOW (-1)
#define FSYS_LOG_SYSLOG	0
#define FSYS_LOG_STDERR	1
#define FSYS_LOG_STDOUT	2
#define FSYS_LOG_FILE	3

/************************************ MACRO DEFINITIONS ***************************************/
#define fsys_divideplus(A, B)	(((A) / (B)) + ((A) % (B) ?1 :0))

/************************************ DATA TYPE DEFINITIONS ***********************************/
/*p_block define a generic reference to a sigle block of the filesystem.
Obviously size of filesystem depend by block size and by u_int type defined in machine.*/

typedef u_int 	p_block;
typedef u_int	p_dentry;

/*directory entry.It must fit in a sigle block.*/
struct dentry{
	p_block         first;  /*block in which the content of file is stored*/
	struct stat     stat;   /*system stat structure*/
	char            name[FSYS_MAXNAMELENGHT +1];    /*name of the file or dir*/
};

/*sblock define the superblock structure, it will be stored at the beginning of filesystem.*/
struct sblock{
	size_t          blocksize;      /*size of block, by mkfs we allow only 1024, 2048 or 4096*/
	size_t          partsize;       /*size of whole filesystem*/
	u_int           blockscount;    /*total block number*/
	size_t          fatsize;        /*lenght of FAT array*/
	size_t          fatblocks;      /*blocks reserved by FAT*/
	p_block         firstfree;      /*pointer to the head of block-free list*/
	struct  dentry  rootdir;        /*root directory entry (description)*/
	char            label[8];       /*name or label of the filesystem*/
};

struct fhandle{
	struct dentry *father;
	struct dentry *gfather;
	u_int offset;
	u_int goffset;
};

/*
struct stat: see man 2 stat
*/

/************************************ GLOBAL VARIABLES DEFINITIONS ****************************/
/*
static int fd;
static struct sblock sb;
static pthread_mutex_t fdmux,fatmux;
*/
int fd;
struct sblock sb;
pthread_mutex_t fdmux,fatmux;


/*for example 4 bytes * 512000 indexes(2MB) in the fat block matches 512000*blocksize(512000*4096=2GB) in the data space*/


/************************************ FUNCTIONS DEFINITIONS ***********************************/

/*ioctl.c*/
int 	fsys_open(char *filename);
int   	fsys_sblockget(int fd,struct sblock *sb);
u_int 	fsys_getnextblock(int fd,struct sblock *sb,u_int index);
u_int 	fsys_getlastblock(int fd,struct sblock *sb,u_int index);
u_int 	fsys_blocknew(int fd,struct sblock *sb);
u_int 	fsys_blockappend(int fd,struct sblock *sb,u_int blocktoappendto);
int   	fsys_blockfree(int fd,struct sblock *sb,p_block index);
int 	fsys_fatwrite(int fd,struct sblock *sb,u_int index,u_int data);
u_int 	fsys_fatread(int fd,struct sblock *sb,u_int index);
int 	fsys_blockread(int fd,struct sblock *sb,u_int block,void *buf,u_int offset, size_t size);
int 	fsys_blockwrite(int fd,struct sblock *sb,u_int block,void *buf,u_int offset, size_t size);
int 	fsys_checkbounds(struct sblock *sb,p_block index);
int 	fsys_close(int fd);
#ifdef FATMAPPED
int 	fsys_fatload(int fd,struct sblock *sb);
int 	fsys_fatsync(int fd,struct sblock *sb);
#endif


/*helper.c*/
int fsys_sblocksync(int fd,struct sblock *sb);
int fsys_getchilddentries(int fd,struct sblock *sb,struct dentry *father,struct dentry *dentries);
int fsys_dentryread(int fd,struct sblock *sb,struct dentry *file,void *buf,u_int offset,size_t size);
int fsys_dentrywrite(int fd,struct sblock *sb,struct dentry *file,void *buf,u_int offset,size_t size);
int fsys_dentrytruncate(int fd,struct sblock *sb,struct dentry *dirent,size_t newsize);
int fsys_dentryblockadd(int fd,struct sblock *sb,struct dentry *dirent);
int fsys_dentryadd(int fd,struct sblock *sb,struct dentry *father,struct dentry *dirent);
int fsys_dentrydel(int fd,struct sblock *sb,struct dentry *father,char *name,int mode);
int fsys_dentrylookup(int fd,struct sblock *sb,struct dentry *father,char *entryname,u_int *offset);
int fsys_lookupfile(int fd,struct sblock *sb,char *fullpath,struct dentry **gfather,u_int *goffset,struct dentry 		**father,u_int *offset);


/*failutils.c*/
struct dentry 	*fsys_init_dentry(mode_t modt,dev_t devt,char *name,int type);
int 		fsys_stringsplit(const char *fullpath,char **fatherpath,char **path);
int 		fsys_aclcheck(struct dentry *dir,int flags);
char 		**fsys_addfuseparam(int argc,char **argv,int optargc, ...);
void 		fsys_printlog(int out,char *frm,...);
void		fsys_logsopen();
void		fsys_logsclose();
void		fsys_usage(void);


#endif
