#include <failsystem.h>

FILE *logfd;
/*extern struct sblock sb;*/

/*initialize a directory entry from name and type of file*/
struct dentry *fsys_init_dentry(mode_t modt,dev_t devt,char *name,int type){
struct dentry *dirent;
struct fuse_context *fcntx;

	dirent=(struct dentry*)malloc(sizeof(struct dentry));
	memset(dirent,0,sizeof(struct dentry));
	if (devt != 0){	/*devt is ignored ! @TODO remove*/
		dirent->stat.st_dev=devt;
	}

	dirent->stat.st_mode=modt | (type<<12);	/* first 12 bit are reserved for permission*/

	/*get uid and guid of current thread operation*/
	fcntx=fuse_get_context();
	dirent->stat.st_uid=fcntx->uid;
	dirent->stat.st_gid=fcntx->gid;
	
	dirent->stat.st_blksize=sb.blocksize;
	dirent->stat.st_atime=dirent->stat.st_mtime=dirent->stat.st_ctime=time(NULL);
	strcpy(dirent->name,name);

	return dirent;
}

/*	return the name and basename from a full path
 *
 *	example: fullpath:/a/b/c --> fatherpath:/a/b	path:c
 *	example: fullpath:/a/b	 --> fatherpath:/a	path:b
 *	example: fullpath:/a	 --> fatherpath:/	path:a
 */
int fsys_stringsplit(const char *fullpath,char **fatherpath,char **path){
char c=1;
u_int slashpos,i;
	i=0;
	while(c!='\0'){
		c=fullpath[i];
		if (fullpath[i]=='/')slashpos=i;
		i++;
	}
	if (slashpos==0){
		*fatherpath=(char*)strndup(fullpath,1);
		*path=(char*)strdup(fullpath+1);
	}
	else{
		*fatherpath=(char*)strndup(fullpath,slashpos);
		*path=(char*)strdup(fullpath+slashpos+1);
	}
	return 0;
}

/*check permission for a file, return 1 for grantend permission otherwise 0*/
int fsys_aclcheck(struct dentry *dir,int flags){
int mod, dirmod;
int ok=0,i;
	/*convert flags of open call to standard *nix permission notation*/
	switch(flags & 0x3){
		case (O_RDONLY):
			mod=4;
			break;
		case (O_RDWR):
			mod=6;
			break;
		
		case (O_WRONLY):
			mod=2;
			break;
	}
	/*get permission of current directory entry*/
	dirmod= (dir->stat.st_mode & 0x1FF);
	
	for(i=0; (i<3 && !ok);i++){
		ok= (mod == (mod & dirmod));
		mod<<=3;
	}
	return ok;
}

/*UNUSED this function add params to fuse library option*/
char **fsys_addfuseparam(int argc,char **argv,int optargc, ...){
va_list args;
char **margv;
int i,mi;

	margv=(char**)malloc(sizeof(char*) * (argc + optargc - 1));
	va_start(args,optargc);
	mi=0;
	margv[mi++]=(char*)strdup(argv[0]);	/*copy the process name*/

	/*arg 2 is not to be passed*/
	for(i=2;i<argc;i++){
		margv[mi++]=(char*)strdup(argv[i]);
	}
	for (i=0;i<optargc;i++){
		margv[mi++]=(char*)strdup( (char*)va_arg(args, char*) );
	}
	va_end(args);
	return margv;
}

/*local enhanced message-print to use syslog and other fds*/
void fsys_printlog(int out,char *frm,...){
va_list args;

	va_start(args,frm);
	switch (out){
		case FSYS_LOG_STDOUT:
			vprintf(frm,args);
			break;
		case FSYS_LOG_SYSLOG:
			vsyslog(LOG_INFO,frm,args);
			break;
		case FSYS_LOG_STDERR:
			vfprintf(stderr,frm,args);
			break;
		case FSYS_LOG_FILE:
			vfprintf(logfd,frm,args);
			break;
	}
	va_end(args);
}

/*open stuff for logging*/
void fsys_logsopen(){
char *filename,*basename,*path;
int namelenght;	
	basename=getenv("HOME");
	filename=strdup("/failsystem.log");
	namelenght=strlen(basename) + strlen(filename);
	
	path=(char*)malloc(namelenght + 1);
	strcpy(path,basename);
	strcat(path,filename);	/*overwrite null-char of path and re-add it at the end*/
	
	logfd=fopen(path,"w+");
	openlog("FailSystem",LOG_PID,LOG_USER);
	free(filename);
	free(path);	
}

/*close stuff for logging*/
void fsys_logsclose(){
	fclose(logfd);
	closelog();
}

/* this comment is a line leek :) */
void fsys_usage(void){
	printf("usage:\n");
	printf("failsystem <filename> <mountpoint> [fuse options]\n");
	printf("NOTE: -s fuse's option is implicit\n");
}

