#include <failsystem.h>
/*
static int fd;
static struct sblock sb;
*/
static p_block *freelist;
static u_int freeblocks;
static int isfreelistwrong=0;
static int areblockswrong=0;
static u_int progressblock=0;

int recurse_dir(struct dentry *dir);
int checkblock(struct dentry *dirent);
u_int getfreeblocks(p_block **list);
void smilebar(u_int max,u_int current);


void usage(void){
	printf("usage:\n");
	printf("fsck.fail  <filename>\n");
}

int main(int argc,char **argv){

	if (argc != 2 ){
		usage();
		exit(1);
	}
	
	if ((fd=fsys_open(argv[1])) < 0){
		perror("");
		exit(1);
	}
	
	fsys_sblockget(fd,&sb);
#ifdef FATMAPPED
	fsys_fatload(fd,&sb);
#endif	

	printf("Phase 1: Checking free list ************************\n");
	freeblocks=getfreeblocks(&freelist);	
	printf("\tFree blocks are: %u\n\n",freeblocks);
	printf("Phase 2: Checking FAT consistence ******************\n");
	recurse_dir(&sb.rootdir);
	
	printf("Free list are wrong ? %d\n",isfreelistwrong);
	printf("Blocks are wrong ? %d\n",areblockswrong);
	if (isfreelistwrong || areblockswrong){
		printf("FAIL-system is DAMAGED !!\n");
	}
	else{
		printf("FAIL-system is OK !!\n");
	}
	free(freelist);
	return 0;
}


int recurse_dir(struct dentry *dir){
struct dentry *direntries;
u_int entriescount,i;
	/*dir is a directory*/
	direntries=(struct dentry*)malloc(dir->stat.st_size);
	fsys_dentryread(fd,&sb,dir,(void*)direntries,0,dir->stat.st_size);
	entriescount=dir->stat.st_size / sizeof(struct dentry);
	printf("Entering directory \"%s\"\this entries are %u\n",dir->name,entriescount);
	for (i=0;i < entriescount;i++){
		if (direntries[i].stat.st_mode & (FSYS_TYPE_DIR<<12)){
			/*is a directory*/
			checkblock(&direntries[i]);
			recurse_dir(&direntries[i]);
		}
		else{
			/*is a file*/
			checkblock(&direntries[i]);
		}
	}
	free(direntries);
	printf("Exiting directory \"%s\"\n",dir->name);
	return 0;
}

int checkblock(struct dentry *dirent){
p_block current;
u_int i,c,blockscount;
	printf("Checking file \"%s\"\n",dirent->name);
	progressblock++;
	smilebar(sb.blockscount -1 - freeblocks,progressblock+1);
	current=dirent->first;
	blockscount=fsys_divideplus(dirent->stat.st_size,sb.blocksize);
	for(c=0;c < blockscount;c++){
/*		printf("\tblock: %u\n",current);	*/
/*		smilebar(blockscount,c+1);	*/
		if (!current){
			areblockswrong=1;
			printf("\tFile size exceded blocks\n");
			continue;
		}
		for (i=0;i<freeblocks;i++){
			if(freelist[i] == current){
				areblockswrong=1;
				printf("\tBlock %u owned by file \"%s\" is corrupted\n",
					current,
					dirent->name);
			}
		}
		current=fsys_getnextblock(fd,&sb,current);
	}
	if (current && blockscount){
		/*current must be zero because end of list should occurred*/
		areblockswrong=1;
		printf("\tLast block dont are in bounds limits\n");
	}
	
	return 0;
}

u_int getfreeblocks(p_block **list){
u_int freeblocks=0,i,c;
p_block current;
	current=sb.firstfree;
	/*get the number of free blocks */
	while(current){
		current=fsys_getnextblock(fd,&sb,current);
		freeblocks++;
		if (freeblocks > (sb.blockscount - sb.fatblocks)){
			/*loop detected*/
			isfreelistwrong=1;
			printf("DANGER !! FREE LIST HAVE A LOOP !!\n");
			break;
		}
	}
	*list=(p_block*)malloc(sizeof(p_block) * freeblocks);
	
	current=sb.firstfree;
	for(i=0;i<freeblocks;i++){
/*		printf("Adding %u to freelist\n",current);	*/
		(*list)[i]=current;
		current=fsys_getnextblock(fd,&sb,current);
	}
	
	/*check for multiple instance of blocks in list*/
	for(i=0;i<freeblocks;i++){
		smilebar(freeblocks,i+1);
		for (c=0;c<i;c++){
			if(freelist[i]==freelist[c]){
				/*Detected a multiple instance*/
				printf("Butta via tutto\n");
			}
		}
	}
	return freeblocks;
}

void smilebar(u_int max,u_int current){
u_int percent;
u_int starscount;
u_int refreshunit,i;
u_int maxstars;
	if (!max) printf(" OK\n");
	
	maxstars=50;
	percent=((u_int64_t)current * 100) / max;
	starscount=((u_int64_t)current * maxstars) / max;
	
	refreshunit=max / maxstars;
	if (!refreshunit)refreshunit=1;
	if ((!(current % refreshunit)) || current==max ){
		/*print all bar*/
		fflush(NULL);
		printf("\r%u%%\t",percent);
		for(i=0;i<starscount;i++){
			printf("*");
		}
	}
	if (current == max)
		printf(" OK\n");
}



