#include <failsystem.h>

/*syncronise superblock on the file*/
int fsys_sblocksync(int fd,struct sblock *sb){
	fsys_blockwrite(fd,sb,0,(void*)sb,0,sizeof(struct sblock));
#ifdef FATMAPPED
	fsys_fatsync(fd,sb);
#endif
	return 0;
}

/*read data from a directory entry by given offset and size, this function search also in linked blocks*/
/*WARNING: it is possible to reads a large file at one time, this is highly deprecated for memory leaks*/
int fsys_dentryread(int fd,struct sblock *sb,struct dentry *file,void *buf,u_int offset,size_t size){
u_int blockstoseek,toread,i,blockoffset,bytestofullblock,nblockstoread,bread;
p_block blocktoread;
size_t sizetoread;
char *mbuf;
	
	mbuf=(char*)buf;
	if(file==NULL){
		/*if file is NULL, i write rootdir*/
		*(struct dentry*)buf=sb->rootdir;
		return sizeof(struct dentry);
	}
	
	toread=size;
	if (offset > file->stat.st_size){
		return 0;	/*return 0 bytes read*/
	}
	if ((offset + size) > file->stat.st_size){
		toread=file->stat.st_size - offset;
	}
	
	/*seek block(s) for offset*/
	blockstoseek=offset / sb->blocksize;
	blocktoread=file->first;
	for (i=0;i<blockstoseek;i++){
		blocktoread=fsys_getnextblock(fd,sb,blocktoread);
	}

	blockoffset=offset % sb->blocksize;
	bytestofullblock=sb->blocksize - blockoffset;
	
	sizetoread=(toread > bytestofullblock ?bytestofullblock :toread);
	bread=fsys_blockread(fd,sb,blocktoread,mbuf,blockoffset,sizetoread);
	mbuf+=sizetoread;
	toread-=sizetoread;

	nblockstoread=fsys_divideplus(toread,sb->blocksize);
	for (i=0;i<nblockstoread;i++){
		blocktoread=fsys_getnextblock(fd,sb,blocktoread);
		sizetoread=(toread > sb->blocksize ? sb->blocksize :toread);
		bread+=fsys_blockread(fd,sb,blocktoread,mbuf,0,sizetoread);
		mbuf+=sizetoread;
		toread-=sizetoread;
	}
	return (u_int)((u_int)mbuf - (u_int)buf);
}

/*write data to a directory entry by given offset and size, this function can allocate new blocks*/
int fsys_dentrywrite(int fd,struct sblock *sb,struct dentry *file,void *buf,u_int offset,size_t size){
u_int blockstoseek,towrite,i,blockstoadd,actualblocks,requiredblocks,blockoffset,bytestofullblock,nblockstowrite,ret;
p_block blocktowrite;
size_t sizetowrite;
char *mbuf;
	
	mbuf=(char*)buf;
	
	if(file==NULL){
		/*if file is NULL, i write rootdir, other param are ignored*/
		sb->rootdir=*(struct dentry*)buf;
		return sizeof(struct dentry);
	}
	actualblocks = fsys_divideplus(file->stat.st_size,sb->blocksize);
	requiredblocks=fsys_divideplus((offset + size),sb->blocksize);
	/*add necessary block if i need them*/
	if (requiredblocks > actualblocks){
		blockstoadd=requiredblocks - actualblocks;
		
		for(i=0;i<blockstoadd;i++){
			ret=fsys_dentryblockadd(fd,sb,file);
			if (ret==FSYS_DISKFULL){
				fsys_dentrytruncate(fd,sb,file,file->stat.st_size);
				return 0;
			}
		}
	}
	/*seek block(s) for offset*/
	blockstoseek=offset / sb->blocksize;
	blocktowrite=file->first;
	for (i=0;i<blockstoseek;i++){
		blocktowrite=fsys_getnextblock(fd,sb,blocktowrite);
	}

	towrite=size;
	blockoffset=offset % sb->blocksize;
	bytestofullblock=sb->blocksize - blockoffset;
	
	sizetowrite=(towrite > bytestofullblock ?bytestofullblock :towrite);
	fsys_blockwrite(fd,sb,blocktowrite,mbuf,blockoffset,sizetowrite);
	mbuf+=sizetowrite;
	towrite-=sizetowrite;

	nblockstowrite=fsys_divideplus(towrite,sb->blocksize);
	for (i=0;i<nblockstowrite;i++){
		blocktowrite=fsys_getnextblock(fd,sb,blocktowrite);
		sizetowrite=(towrite > sb->blocksize ? sb->blocksize :towrite);
		fsys_blockwrite(fd,sb,blocktowrite,mbuf,0,sizetowrite);
		mbuf+=sizetowrite;
		towrite-=sizetowrite;
	}
	return size;
}

/* search for a direntry by name in the given father direntry and fit offset of it self in the father 
 * return FSYS_FILEMATCH if filename is match, otherwise FSYS_FILENOMATCH
 */
int fsys_dentrylookup(int fd,struct sblock *sb,struct dentry *father,char *entryname,u_int *offset){
struct dentry tempdir;
u_int brotherscount,i;
	brotherscount=father->stat.st_size / (sizeof(struct dentry));
	for (i=0;i<brotherscount;i++){
		fsys_dentryread(fd,sb,father,&tempdir,i*sizeof(struct dentry),sizeof(struct dentry));
		if (!strcmp(entryname,tempdir.name)){
			/*is equal*/
			*offset=i*sizeof(struct dentry);
			return FSYS_FILEMATCH;
		}
	}
	return FSYS_FILENOMATCH;
}

/* lookup a direntry by full path and fit father and gfather respectively with their relative offset
*  it return FSYS_FILEMATCH if full path exists otherwise FSYS_FILENOMATCH
*/
int fsys_lookupfile(int fd,struct sblock *sb,char *fullpath,struct dentry **gfather,u_int *goffset,struct dentry **father,u_int *offset){
char c=1;
int i;
int start=1;
char *fname;
struct dentry currentdir,tmpdir;
u_int tmpoffset;

	*father=NULL;
	if(gfather!=NULL)*gfather=NULL;
	if (!strcmp(fullpath,"/")){return FSYS_FILEMATCH;}
	currentdir=sb->rootdir;
	
	for (i=1;i<=strlen(fullpath);i++){
		c=fullpath[i];
		if (c == '/' || c == '\0'){
			fname=(char*)strndup(fullpath+start,i-start);
			fsys_printlog(FSYS_LOG_SYSLOG,"fsys::lookup filename %s\n",fname);
			/*@TODO*/
			if (fsys_dentrylookup(fd,sb,&currentdir,fname,&tmpoffset) == FSYS_FILENOMATCH){
				if (*father!=NULL) {free(*father); *father=NULL;}
				if (gfather!=NULL && *gfather!=NULL) {free(*gfather); *gfather=NULL;}
				
				free(fname);
				return FSYS_FILENOMATCH;
			}
			else{
				/*name match*/
				if(*father==NULL){
					*father=(struct dentry*)malloc(sizeof(struct dentry));
				}
				**father=currentdir;
				*offset=tmpoffset;
		
				if (c=='\0'){
					free(fname);
					return FSYS_FILEMATCH;
				}
				if(gfather!=NULL){	
					if (*gfather==NULL){
						*gfather=(struct dentry*)malloc(sizeof(struct dentry));
					}
					**gfather=currentdir;
				}
				if(goffset!=NULL){	*goffset=tmpoffset;}
			}
			fsys_dentryread(fd,sb,&currentdir,&tmpdir,tmpoffset,sizeof(struct dentry));
			currentdir=tmpdir;
			
			/*END TODO*/
			free(fname);
			i++;
			start=i;
		}
	}
	
	fsys_printlog(FSYS_LOG_SYSLOG,"lookup assert failed\n");
	free(fname);/*insanity check*/
	return FSYS_FILENOMATCH;
}

/*free blocks from direntry by new specified size, it modifies internal size field of direntry*/
int fsys_dentrytruncate(int fd,struct sblock *sb,struct dentry *dirent,size_t newsize){
p_block currentblock,lastblock;
u_int i,neededblockscount;
	/*@TODO newsize must be less of dirent.size*/
	currentblock=dirent->first;
	if (!currentblock){
		return 0;/*already empty*/
	}
	neededblockscount=fsys_divideplus(newsize,sb->blocksize);/*new blocks count*/
	for (i=0;i<neededblockscount;i++){
		lastblock=currentblock;
		currentblock=fsys_getnextblock(fd,sb,currentblock);
	}
	
	fsys_fatwrite(fd,sb,lastblock,FSYS_BUSY);
	
	fsys_blockfree(fd,sb,currentblock);/*if currentblock is zero, nothing occurs*/
	if (!neededblockscount) dirent->first=0;
	dirent->stat.st_size=newsize;
	dirent->stat.st_blocks=neededblockscount;
	return 0;
}

/*allocs one block to a direntry*/
int fsys_dentryblockadd(int fd,struct sblock *sb,struct dentry *dirent){
u_int last,ret;
	if (!dirent->first){/*or dirent->stat.st_size=0*/
		/*direntry dont has a block, attach a new block*/
		ret=fsys_blocknew(fd,sb);
		if (ret==FSYS_DISKFULL) return ret;
		dirent->first=ret;
	}
	else{
		/*need to append block(s)*/
		last=fsys_getlastblock(fd,sb,dirent->first);
		ret=fsys_blockappend(fd,sb,last);
		if (ret==FSYS_DISKFULL) return ret;
	}
	return 1;
	/*return ret and remove checks above ?*/
}

/*add a direntry to father*/
int fsys_dentryadd(int fd,struct sblock *sb,struct dentry *father,struct dentry *dirent){
u_int internaloffset,ret;
	internaloffset=father->stat.st_size;
	ret=fsys_dentrywrite(fd,sb,father,dirent,internaloffset,sizeof(struct dentry));
	if(ret==FSYS_DISKFULL){
		return FSYS_DISKFULL;
	}
	father->stat.st_size+=sizeof(struct dentry);
	father->stat.st_blocks=fsys_divideplus(father->stat.st_size,sb->blocksize);
	father->stat.st_mtime=time(NULL);
	father->stat.st_ctime=time(NULL);

	return 1;
}

/*delete a direntry from father, it can works in 2 ways : if mode is FSYS_TRALL it deletes directory entry and its content
* if mode is FSYS_TRDENTRY, it deletes only the directory entry.
*/
int fsys_dentrydel(int fd,struct sblock *sb,struct dentry *father,char *name,int mode){
u_int internaloffset,brotherscount;
struct dentry lastdentry,dir;
	/*@TODO error handler */
	fsys_dentrylookup(fd,sb,father,name,&internaloffset);
	fsys_dentryread(fd,sb,father,(void*)&dir,internaloffset,sizeof(struct dentry));
	if (mode)fsys_dentrytruncate(fd,sb,&dir,0);
	brotherscount=father->stat.st_size / sizeof(struct dentry);
	if(brotherscount > 1){
		/*need to defrag father*/
		/*get last brother*/
		fsys_dentryread(fd,sb,father,&lastdentry,(father->stat.st_size - sizeof(struct dentry)),sizeof(struct dentry));
		/*overwrite*/
		fsys_dentrywrite(fd,sb,father,(void*)&lastdentry,internaloffset,sizeof(struct dentry));
	}

	fsys_dentrytruncate(fd,sb,father,(father->stat.st_size - sizeof(struct dentry)));
	return 0;
}
