#include <failsystem.h>

void usage(void){
	printf("usage:\n");
	printf("mkfs_fail <filename> <fs-size> [<blocksize>]\n");
	printf("fs-size is in KB\n");
	printf("allowed blocksize are: 1024 - 2048 - 4096\n");
	printf("if not specified blocksize is set to 1024\n\n");
}

int main(int argc,char **argv){
char zero=0;
int fd, written;
u_int *fat,freespacedata, bsize=0, i, pos, toseek;
struct sblock *sb;	

	/* check command line */
	if (argc < 3 || argc > 4){
		usage();
		return 0;
	}
	if ( argc == 3){
		bsize=1024;
	}
	if ( argc == 4){
		bsize=atoi(argv[3]);
/*		if ((bsize % 1024) || !bsize)	*/
		if( (bsize != 1024) && (bsize != 2048) && (bsize != 4096)){
			printf("Wrong block size!\n");
			usage();
			return 0;
		}
	}
	
	/*creating superblock*/
	sb=(struct sblock*)malloc(sizeof(struct sblock));
	sb->blocksize=bsize;
	sb->partsize=(u_int)((int)atoi(argv[2])<<10);
	sb->blockscount=(u_int)(sb->partsize / sb->blocksize );
	/* normalize file dimension to a multiple of blocksize */
	sb->partsize=(u_int)(sb->blockscount * sb->blocksize);
	sb->fatsize=(size_t)sb->blockscount*sizeof(u_int);
	sb->fatblocks=(u_int)fsys_divideplus(sb->fatsize, sb->blocksize);
	sb->firstfree=(p_block)sb->fatblocks + 1;
	
	printf("Total size:\t\t%u bytes (%u KB)\n",sb->partsize, sb->partsize>>10);
	freespacedata= (sb->blockscount - sb->fatblocks) * sb->blocksize ;
	printf("Free space for data is: %u bytes (%u KB)\n",(u_int)freespacedata,(u_int)freespacedata>>10);
	printf("Blocksize:\t\t%u bytes\n",sb->blocksize);	
	printf("Blocks number:\t\t%u\n",sb->blockscount);
	printf("FAT size:\t\t%u bytes\n",sb->fatsize);
	printf("FAT blocks:\t\t%u\n",sb->fatblocks);

	sb->rootdir.stat.st_mode=0777 | (FSYS_TYPE_DIR<<12);	/* first 12 bit are reserved for permission*/
	sb->rootdir.stat.st_uid=getuid();
	sb->rootdir.stat.st_gid=getgid();
	
	sb->rootdir.stat.st_blksize=sb->blocksize;
	sb->rootdir.stat.st_atime=sb->rootdir.stat.st_mtime=sb->rootdir.stat.st_ctime=time(NULL);
	strcpy(sb->rootdir.name,"/");
	strcpy(sb->label,"fail");
	
	/*creating and initializing FAT*/
	fat=(u_int*)malloc(sb->fatsize);	
	/*fat[0] is superblock,fat[1] to fat[fatblocks] is fat*/
	fat[0]=FSYS_RESERVED;
	for(i=1; i< (sb->fatblocks + 1);i++){
		fat[i]=FSYS_RESERVED;
	}
	/* init free blocks list */
	for(; i< (sb->blockscount-1);i++){
		fat[i]=i+1;
	}
	fat[i]=FSYS_BUSY;/*i indicate the tail of the list */

	/* writing to disk */
/*	sprintf(filename,"%s",argv[1]);	*/
	
	fd=open(argv[1],O_RDWR | O_CREAT ,0660);
	if (fd == -1){
		perror("open failed\nAborting");
		return -1;
	}
	
	/*write superblock and seek until the end of first block*/
	written=write(fd,(void*)sb,sizeof(struct sblock));
	if (written != sizeof(struct sblock)){
		perror("superblock");
		return -1;
	}
	toseek= sb->blocksize - sizeof(struct sblock);
	pos=lseek(fd,toseek,SEEK_CUR);

	/*write fat*/
	written=write(fd,(void*)fat,sb->fatsize);
	if (written != sb->fatsize){
		perror("fat");
		return -1;
	}
	toseek= sb->blocksize - (sb->fatsize % sb->blocksize);
	pos=lseek(fd,toseek, SEEK_CUR);
	
	/* writing end of partition */
	lseek(fd,sb->partsize -1,SEEK_SET);
	write(fd,(void*)&zero,1);
	close(fd);
	printf("Written %d bytes for filesystem\n",pos);
	
	free(sb);
	free(fat);
	printf("Fail-system succesfully created!\n");
	return 0;
}

